import numpy as np 
from tensor import Tensor
from graph import draw_dot
from micrograd.engine import Value
import torch


# class Linear:
#     def __init__(self, in_features: int, out_features: int, bias: bool = True):
#         self.weight = Tensor.normal(0, 1, (out_features, in_features))
#         self.bias = Tensor.normal(0, 1, (out_features,)) if bias else None
    
#     def __call__(self, x: Tensor) -> Tensor:
#         out = x @ self.weight.T  # Matrix multiplication
#         if self.bias is not None:
#             out += self.bias
#         return out


# a = Tensor.ones((2,3), requires_grad=True)
# b = Tensor.full((3,2), 3., requires_grad=True)
# c = a @ b
# loss = c.sum()
# loss.backward()
# print(a.grad)
# print(b.grad)

# draw_dot(loss, filename="img/graph_inspect", inspect=True)

# a = torch.ones((2,3), requires_grad=True)
# b = torch.full((3,2), 3., requires_grad=True)
# c = a @ b
# loss = c.sum()
# loss.backward()
# print(a.grad)
# print(b.grad)


# a = Tensor.ones((3,2), requires_grad=True)
# aa = a.reshape((2,3))
# b = Tensor.full((3,2), 3., requires_grad=True)
# c = aa @ b
# loss = c.sum()
# loss.backward()
# print(a.grad)
# print(b.grad)
# print(aa.grad)
# draw_dot(loss, filename="img/graph_inspect", inspect=True)

# a = torch.ones((3,2), requires_grad=True)
# aa = a.view((2,3))
# b = torch.full((3,2), 3., requires_grad=True)
# c = aa @ b
# loss = c.sum()
# loss.backward()
# print(a.grad)
# print(b.grad)

# a = Tensor.ones((3,3), requires_grad=True)
# aa = a[0:2]
# b = Tensor.full((3,2), 3., requires_grad=True)
# c = aa @ b
# loss = c.sum()
# loss.backward()
# draw_dot(loss, filename="img/graph_inspect", inspect=True)
# print(a.grad)
# print(b.grad)

# a = torch.ones((3,3), requires_grad=True)
# aa = a[0:2]
# b = torch.full((3,2), 3., requires_grad=True)
# c = aa @ b
# loss = c.sum()
# loss.backward()
# print(a.grad)
# print(b.grad)

# a = Tensor.eye(3, requires_grad=True)
# b = Tensor.full((3,3), 3., requires_grad=True)
# c = a @ b
# d = c.mean(axis=0, keepdim=True)
# e = Tensor.full((1,3), 4., requires_grad=True)
# f = d * e
# loss = f.sum()
# loss.backward()
# res_robin_1 = a.grad.tolist()
# print(res_robin_1)
# # torch 
# a = torch.eye(3, requires_grad=True)
# b = torch.full((3,3), 3., requires_grad=True)
# c = a @ b
# d = c.mean(axis=0, keepdim=True)
# e = torch.full((1,3), 4., requires_grad=True)
# f = d * e
# loss = f.sum()
# loss.backward()
# res_torch_1 = a.grad.numpy().tolist()
# print(res_torch_1)

# a = Tensor.full((5,1), 3., requires_grad=True)
# b = Tensor.full((1,1), 2., requires_grad=True)
# c = a + b
# loss = c.mean()
# try:
#     loss.backward()
#     print(loss.grad)
#     print(c.grad)
#     print(a.grad)
#     print(b.grad)
#     draw_dot(loss, filename="img/graph_inspect", inspect=True)
# except Exception as e:
#     print(e)

# print("\n")
# print("- TORCH -")
# a = torch.full((5,1), 3., requires_grad=True)
# b = torch.full((1,1), 2., requires_grad=True)
# c = a + b
# loss = c.mean()
# try:
#     loss.backward()
#     print("torch win")
#     print(a.grad)
#     print(b.grad)
# except Exception as e:
#     print(e)



# a = Tensor.full((5,1), 3., requires_grad=True)
# b = Tensor.ones((1,), requires_grad=True)
# c = b + a
# print(c.shape)
# loss = c.sum()
# try:
#     loss.backward()
# except Exception as e:
#     print(e)
# print(a.grad)
# print(a.shape, a.grad.shape)
# print(b.grad)
# print(b.shape, b.grad.shape)

# print('\nTORCH')
# a = torch.full((5,1), 3., requires_grad=True)
# b = torch.ones((1,), requires_grad=True)
# c = b + a
# loss = c.sum()
# loss.backward()
# print(a.grad)
# print(a.shape, a.grad.shape)
# print(b.grad)
# print(b.shape, b.grad.shape)



# a = Tensor.full((5,1), 3., requires_grad=True)
# b = Tensor.ones((1,), requires_grad=True)
# c = a * b
# loss = c.sum()
# try:
#     loss.backward()
# except Exception as e:
#     print(e)
# print(a.grad)
# # print(a.shape, a.grad.shape)
# print(b.grad)
# # print(b.shape, b.grad.shape)

# print('\nTORCH')
# a = torch.full((5,1), 3., requires_grad=True)
# b = torch.ones((1,), requires_grad=True)
# c = a * b
# loss = c.sum()
# loss.backward()
# print(a.grad)
# # print(a.shape, a.grad.shape)
# print(b.grad)
# # print(b.shape, b.grad.shape)


# print("\nREVERSE\n")

# a = Tensor.full((5,1), 3., requires_grad=True)
# b = Tensor.ones((1,), requires_grad=True)
# c = b * a
# loss = c.sum()
# try:
#     loss.backward()
# except Exception as e:
#     print(e)
# print(a.grad)
# # print(a.shape, a.grad.shape)
# print(b.grad)
# # print(b.shape, b.grad.shape)

# print('\nTORCH')
# a = torch.full((5,1), 3., requires_grad=True)
# b = torch.ones((1,), requires_grad=True)
# c = b * a
# loss = c.sum()
# loss.backward()
# print(a.grad)
# # print(a.shape, a.grad.shape)
# print(b.grad)
# # print(b.shape, b.grad.shape)

# a = Value(-4.0)
# b = Value(2.0)
# c = a + b
# d = a * b + b**3
# c += c + 1
# c += 1 + c + (-a)
# d += d * 2 + (b + a).relu()
# d += 3 * d + (b - a).relu()
# e = c - d
# f = e**2
# g = f / 2.0
# g += 10.0 / f
# print(f'{g.data:.4f}') # prints 24.7041, the outcome of this forward pass
# g.backward()
# print(f'{a.grad:.4f}') # prints 138.8338, i.e. the numerical value of dg/da
# print(f'{b.grad:.4f}') # prints 645.5773, i.e. the numerical value of dg/db

# print("\nrobin\n")
# a = Tensor(-4.0, requires_grad=True)
# b = Tensor(2.0, requires_grad=True)
# c = a + b
# d = a * b + b**3
# c += c + 1
# c += 1 + c + (-a)
# d += d * 2 + (b + a).relu()
# d += 3 * d + (b - a).relu()
# e = c - d
# f = e**2
# g = f / 2.0
# g += 10.0 / f
# print(f'{g.data:.4f}') # prints 24.7041, the outcome of this forward pass
# g.backward()
# print(f'{a.grad:.4f}') # prints 138.8338, i.e. the numerical value of dg/da
# print(f'{b.grad:.4f}') # prints 645.5773, i.e. the numerical value of dg/db



a = Tensor.full((5,1), 3., requires_grad=True)
b = Tensor.ones((1,1), requires_grad=True)
c = b * a
loss = c.sum()
try:
    loss.backward()
except Exception as e:
    print(e)
print(a.grad)
print(b.grad)

a = torch.full((5,1), 3., requires_grad=True)
b = torch.ones((1,1), requires_grad=True)
c = b * a
loss = c.sum()
loss.backward()
print(a.grad)
print(b.grad)
# res_torch_1 = a.grad.numpy().tolist()
# res_torch_2 = b.grad.numpy().tolist()

# print(res_robin_1)
# print(res_robin_2)
# print('\nTORCH')
# print(res_torch_1)
# print(res_torch_2)


from robingrad.tensor import Tensor
from .loss import MSELoss

class Linear:
    def __init__(self, in_features: int, out_features: int, bias: bool = True):
        self.weight = Tensor.normal(0,1,(out_features, in_features), requires_grad=True)
        self.bias = Tensor.normal(0,1,(1,out_features), requires_grad=True) if bias else None
    
    def __call__(self, x: "Tensor") -> "Tensor":
        return x.linear(self.weight.T, self.bias)
    


from robingrad.tensor import Tensor
import numpy as np
from typing import List

class Optimizer:
    def __init__(self, params: List[Tensor]):
        self.params = params

    def zero_grad(self) -> None:
        for param in self.params: 
            param.grad = np.zeros_like(param.grad)
        

class SGD(Optimizer):
    def __init__(self, params: List[Tensor], lr: float = 3e-4):
        super().__init__(params)
        self.lr = lr

    def step(self) -> None:
        for param in self.params:
            param.data += -self.lr * param.grad


# https://pytorch.org/docs/stable/generated/torch.optim.Adam.html
class Adam(Optimizer):
    def __init__(self, params: List[Tensor], lr: float = 0.001, beta1: float = 0.9, beta2: float =  0.999, eps: float = 1e-08):
        super().__init__(params)
        self.lr, self.beta1, self.beta2, self.eps = lr, beta1, beta2, eps
        self.m = [np.zeros_like(param.data) for param in self.params] # first moment
        self.v = [np.zeros_like(param.data) for param in self.params] # second moment 
        self.t = 0

    def step(self) -> None:
        self.t += 1
        for i, param in enumerate(self.params):
            self.m[i] = self.beta1*self.m[i] + (1-self.beta1)*param.grad
            self.v[i] = self.beta2*self.v[i] + (1-self.beta2)*(param.grad**2)
            m_hat = self.m[i] / (1-self.beta1**self.t) # first moment corrected
            v_hat = self.v[i] / (1-self.beta2**self.t) # second moment corrected
            param.data -= (self.lr * m_hat)/(np.sqrt(v_hat) + self.eps)
            



from robingrad.tensor import Tensor
from typing import List, Union

class MSELoss:
    def __call__(self, output: Union[List[Tensor], Tensor], target: Union[List[Tensor], Tensor]) -> Tensor:
        return ((output - target)**2).mean()
        


from graphviz import Digraph
import numpy as np


def trace(root):
    nodes, edges = set(), set()
    def build(v):
        if v not in nodes:
            nodes.add(v)
            for child in v._prev:
                edges.add((child, v))
                build(child)
    build(root)
    return nodes, edges


def draw_dot(root, filename: str = None, inspect: bool = False, format: str = 'svg', rankdir: str = 'LR'):
    """
    filename: name of the output file to save the DOT file (without extension)
    inspect: flag to include detailed information in the labels
    format: png | svg | ...
    rankdir: TB (top to bottom graph) | LR (left to right)
    """
    assert rankdir in ['LR', 'TB']
    nodes, edges = trace(root)
    dot = Digraph(format=format, graph_attr={'rankdir': rankdir})

    for n in nodes:
        if inspect:
            dot.node(name=str(id(n)), label=f"data: {np.round(n.data,4)} | grad: {np.round(n.grad,4)} | shape: {n.shape} | type: {n.data.dtype}", shape='record')
        else:
            dot.node(name=str(id(n)), label=f"name: {n._origin} | shape: {n.shape} | type: {n.data.dtype} | grad: {n.requires_grad}", shape='record')
        if n._op:
            dot.node(name=str(id(n)) + n._op, label=n._op)
            dot.edge(str(id(n)) + n._op, str(id(n)))

    for n1, n2 in edges:
        dot.edge(str(id(n1)), str(id(n2)) + n2._op)

    if filename:
        dot.render(filename, format="png",view=True) 

    return dot


import numpy as np
from typing import Tuple, Union, List, TypeVar, Optional

DType = TypeVar("DType", bound=np.dtype)

# Trust in broadcasting rules
class Tensor:
    def __init__(self, data: Union[np.ndarray, List, int, float], dtype: DType = np.float32, requires_grad: bool = False, _children: Tuple =(), _op: str = "", _origin: str = "tensor"):
        self.data = np.asarray(data).astype(dtype)
        if requires_grad:
            self.grad = np.zeros_like(data).astype(dtype)
        else:
            self.grad = None
        self._backward = lambda: None
        self._prev = set(_children)
        self._op = _op
        self._origin = _origin 
        self.requires_grad = requires_grad
    
    @staticmethod
    def zeros(shape: Tuple, **kwargs) -> "Tensor": return Tensor(np.zeros(shape), _origin="zeros", **kwargs)

    @staticmethod
    def zeros_like(tensor: "Tensor", **kwargs) -> "Tensor": return Tensor(np.zeros_like(tensor.data), _origin="zeros_like", **kwargs)

    @staticmethod
    def eye(size: int, **kwargs) -> "Tensor": return Tensor(np.eye(size), _origin="eye", **kwargs)

    @staticmethod
    def full(shape: Tuple, fill_value: Union[float, int], **kwargs) -> "Tensor": return Tensor(np.full(shape,fill_value), _origin="full", **kwargs)

    @staticmethod
    def full_like(tensor: "Tensor", fill_value: Union[float, int], **kwargs) -> "Tensor": return Tensor(np.full_like(tensor.data, fill_value), _origin="full like", **kwargs)

    @staticmethod
    def ones(shape: Tuple, **kwargs) -> "Tensor": return Tensor(np.ones(shape), _origin="ones", **kwargs)

    @staticmethod
    def ones_like(tensor: "Tensor", **kwargs) -> "Tensor": return Tensor(np.ones_like(tensor.data), _origin="ones like", **kwargs)
    
    @staticmethod
    def broadcast(reference: "Tensor", target: "Tensor", **kwargs) -> "Tensor": 
        if np.broadcast(reference, target).shape == np.broadcast(reference, reference).shape:
            return Tensor(np.broadcast_to(target, reference.shape), _origin="broadcasted", **kwargs)
        else:
            raise ValueError("Tensors are not broadcastable.")
        
    @staticmethod
    def normal(mean: Union[int, float], std: Union[int, float], shape: Tuple, **kwargs) -> "Tensor": return Tensor(np.random.normal(mean, std, shape), _origin="normal", **kwargs)
        
    @staticmethod
    def uniform(low: Union[int, float], high: Union[int, float], shape: Tuple, **kwargs) -> "Tensor": return Tensor(np.random.uniform(low, high, shape), _origin="uniform", **kwargs)    

    def sum(self) -> "Tensor":
        out = Tensor(self.data.sum(), dtype=self.data.dtype, _children=(self,), _op="sum()", _origin="sum", requires_grad=self.requires_grad)
        
        def _backward():
            self.grad += np.ones_like(self.data) * out.grad
        out._backward = _backward
        
        return out
    
    def mean(self, axis=None, keepdim: bool = False) -> "Tensor":
        if axis is None:
            data = self.data.mean()
        else:
            data = self.data.mean(axis=axis, keepdims=keepdim)
        out = Tensor(data, dtype=self.data.dtype, _children=(self,), _op="μ", _origin="mean", requires_grad=self.requires_grad)
    
        def _backward():
            if axis is None:
                self.grad += np.ones_like(self.data) * out.grad / self.data.size
            else:
                axis_sum_size = self.data.shape[axis] if keepdim else 1
                self.grad += np.ones_like(self.data) * out.grad / axis_sum_size
        out._backward = _backward

        return out

    def relu(self) -> "Tensor":
        out = Tensor(np.maximum(0, self.data), dtype=self.data.dtype, _children=(self,), _op="relu()", _origin="ReLU", requires_grad=self.requires_grad)

        def _backward():
            self.grad += (out.data > 0) * out.grad
        out._backward = _backward

        return out
    
    def sigmoid(self) -> "Tensor":
        x = self.data
        t = (1 + np.exp(-x))**-1
        out = Tensor(t, dtype=self.data.dtype, _children=(self,), _op="sigmoid()", _origin="sigmoid", requires_grad=self.requires_grad)
        
        def _backward():
            self.grad += t*(1-t) * out.grad 
        out._backward = _backward

        return out 
    
    def tanh(self) -> "Tensor":
        x = self.data
        t = (np.exp(2*x) - 1)/(np.exp(2*x) + 1)
        out = Tensor(t, dtype=self.data.dtype, _children=(self,), _op="tanh()", _origin="tanh", requires_grad=self.requires_grad)
        
        def _backward():
            self.grad += (1 - t**2) * out.grad
        out._backward = _backward
        
        return out

    def exp(self) -> "Tensor":
        x = self.data
        out = Tensor(np.exp(x), dtype=self.data.dtype, _children=(self,), _op="exp()", _origin="exp", requires_grad=self.requires_grad)
        
        def _backward():
            self.grad += out.data * out.grad 
        out._backward = _backward
        
        return out
    
    def log(self) -> "Tensor":
        if np.any(self.data <= 0):
            raise ValueError("can't log negative or zero value")
        x = self.data
        out = Tensor(np.log(x), dtype=self.data.dtype, _children=(self,), _op="log()", _origin="log", requires_grad=self.requires_grad)

        def _backward():
            self.grad = (x**(-1)) * out.grad
        out._backward = _backward

        return out
    
    def __add__(self, other: Union["Tensor", np.ndarray, List, int, float]) -> "Tensor":
        other = other if isinstance(other, Tensor) else Tensor.broadcast(self, other, dtype=self.data.dtype, requires_grad=self.requires_grad)
        out = Tensor(self.data + other.data, dtype=self.data.dtype, _children=(self, other), _op='+', _origin="__add__", requires_grad=self.requires_grad)
        
        def _backward():
            self.grad += np.sum(out.grad, axis=0) if self.data.shape != out.data.shape else out.grad
            # other.grad += out.grad
            other.grad += np.sum(out.grad, axis=0) if other.data.shape != out.data.shape else out.grad
        out._backward = _backward

        return out
    
    def __mul__(self, other: Union["Tensor", np.ndarray, List, int, float]) -> "Tensor":
        other = other if isinstance(other, Tensor) else Tensor.broadcast(self, other, dtype=self.data.dtype, requires_grad=self.requires_grad)
        out = Tensor(self.data * other.data, dtype=self.data.dtype, _children=(self, other), _op='*', _origin="__mul__", requires_grad=self.requires_grad)

        def _backward():
            # self.grad += other.data * out.grad
            self.grad += np.sum(other.data * out.grad, axis=0) if self.data.shape != out.data.shape else (other.data * out.grad)
            # other.grad += self.data * out.grad
            other.grad += np.sum(self.data * out.grad, axis=0) if other.data.shape != out.data.shape else (self.data * out.grad)
        out._backward = _backward

        return out
    
    def __pow__(self, other: Union[int, float]) -> "Tensor":
        assert isinstance(other, (int, float)), "only supporting int/float powers for now"
        out = Tensor(self.data ** other, dtype=self.data.dtype, _children=(self,), _op=f'**{other}', _origin="__pow__", requires_grad=self.requires_grad)

        def _backward():
            self.grad += (other * self.data**(other-1)) * out.grad
        out._backward = _backward

        return out
    
    def __matmul__(self, other: Union["Tensor", np.ndarray, List, int, float]) -> "Tensor":
        other = other if isinstance(other, Tensor) else Tensor(other, dtype=self.data.dtype, requires_grad=self.requires_grad)
        out = Tensor(self.data @ other.data, dtype=self.data.dtype, _children=(self, other), _op='@', _origin="__matmul__", requires_grad=self.requires_grad)

        def _backward():
            self.grad += out.grad @ other.data.T
            other.grad += self.data.T @ out.grad
        out._backward = _backward

        return out
    
    def __rmatmul__(self, other: Union["Tensor", np.ndarray, List, int, float]) -> "Tensor":
        return self @ other
    
    def __neg__(self) -> "Tensor": # -self
        return self * Tensor.full_like(self, -1, requires_grad=self.requires_grad)

    def __radd__(self, other: Union["Tensor", np.ndarray, List, int, float]) -> "Tensor": # other + self
        return self + other

    def __sub__(self, other: Union["Tensor", np.ndarray, List, int, float]) -> "Tensor": # self - other
        return self + (-other)

    def __rsub__(self, other: Union["Tensor", np.ndarray, List, int, float]) -> "Tensor": # other - self
        return other + (-self)

    def __rmul__(self, other: Union["Tensor", np.ndarray, List, int, float]) -> "Tensor": # other * self
        return self * other

    def __truediv__(self, other: Union["Tensor", np.ndarray, List, int, float]) -> "Tensor": # self / other
        return self * other**-1

    def __rtruediv__(self, other: Union["Tensor", np.ndarray, List, int, float]) -> "Tensor": # other / self
        return other * self**-1
    
    def backward(self) -> None:
        if self.shape != () and self.shape != (1,) and self.shape != (1,1):
            raise ValueError("Backward can only be called on a scalar tensor.")
        # topological order all of the children in the graph
        topo = []
        visited = set()
        def build_topo(v):
            if v not in visited:
                visited.add(v)
                for child in v._prev:
                    build_topo(child)
                topo.append(v)
        build_topo(self)

        # go one variable at a time and apply the chain rule to get its gradient
        self.grad = 1
        for v in reversed(topo):
            v._backward()
            
    @property
    def shape(self) -> Tuple[int, int]: return self.data.shape

    @property
    def dtype(self) -> str: return self.data.dtype
            
    def __repr__(self) -> str:
        if self.requires_grad:
            return f"Tensor: {self._origin}\ndata: \n{self.data}\ngrad: \n{self.grad}\ndtype: {self.data.dtype}"
        else:
            return f"Tensor: {self._origin}\ndata: \n{self.data}\ndtype: {self.data.dtype}"
    
    def __getitem__(self, val) -> "Tensor":
        out = Tensor(self.data[val], dtype=self.data.dtype, _children=(self,), _op="slice", _origin="slice", requires_grad=self.requires_grad)
        
        def _backward():
            self.grad[val] += out.grad
        out._backward = _backward

        return out

    def reshape(self, shape: Tuple[int, int]) -> "Tensor":
        out = Tensor(self.data.reshape(shape), dtype=self.data.dtype, _children=(self,), _op="reshape", _origin="reshape", requires_grad=self.requires_grad)
        
        def _backward():
            self.grad += out.grad.reshape(self.shape)
        out._backward = _backward

        return out
    
    def transpose(self, ax1: int = 1, ax2: int = 0) -> "Tensor":
        out = Tensor(self.data.transpose(ax1, ax2), dtype=self.data.dtype, _children=(self,), _op="T", _origin="T", requires_grad=self.requires_grad)

        def _backward():
            self.grad += out.grad.transpose(ax1, ax2)
        out._backward = _backward

        return out

    @property
    def T(self) -> "Tensor": return self.transpose()  
    
    def linear(self, weight: "Tensor", bias: Optional["Tensor"] = None) -> "Tensor":
        x = self * weight if len(weight.shape) == 1 else self @ weight
        return x + bias if bias is not None else x

    def flatten(self) -> "Tensor":
        out = Tensor(self.data.flatten(), dtype=self.data.dtype, _children=(self,), _op="flatten", _origin="flatten", requires_grad=self.requires_grad)

        def _backward():
            self.grad = out.grad.reshape(self.shape)
        out._backward = _backward

        return out 



from robingrad import Tensor
import torch
import unittest
from micrograd.engine import Value
import numpy as np

class TestTensor(unittest.TestCase):

    def test_tensor_ops(self):
        # robin
        a = Tensor.eye(3, requires_grad=True)
        b = Tensor.full((3,3), 3., requires_grad=True)
        c = Tensor.ones((3,3), requires_grad=True)
        loss = ((a+b-c)**2).sum()
        loss.backward()
        res_robin = a.grad.tolist()
        # torch
        a = torch.eye(3, requires_grad=True)
        b = torch.full((3,3),3.,requires_grad=True)
        c = torch.ones((3,3), requires_grad=True)
        loss = ((a+b-c)**2).sum()
        loss.backward()
        res_torch = a.grad.numpy().tolist()
        # test 
        self.assertAlmostEqual(res_robin, res_torch)

    def test_robin_micrograd(self):
        # robin
        a = Tensor(-4.0, requires_grad=True)
        b = Tensor(2.0, requires_grad=True)
        c = a + b
        d = a * b + b**3
        c += c + 1
        c += 1 + c + (-a)
        d += d * 2 + (b + a).relu()
        d += 3 * d + (b - a).relu()
        e = c - d
        f = e**2
        g = f / 2.0
        g += 10.0 / f
        #print(f'{g.data:.4f}') # prints 24.7041, the outcome of this forward pass
        g.backward()
        res_robin = np.round(float(a.grad),4)
        #print(f'{a.grad:.4f}') # prints 138.8338, i.e. the numerical value of dg/da
        #print(f'{b.grad:.4f}')
        # micrograd
        a = Value(-4.0)
        b = Value(2.0)
        c = a + b
        d = a * b + b**3
        c += c + 1
        c += 1 + c + (-a)
        d += d * 2 + (b + a).relu()
        d += 3 * d + (b - a).relu()
        e = c - d
        f = e**2
        g = f / 2.0
        g += 10.0 / f
        #print(f'{g.data:.4f}') # prints 24.7041, the outcome of this forward pass
        g.backward()
        #print(f'{a.grad:.4f}') # prints 138.8338, i.e. the numerical value of dg/da
        #print(f'{b.grad:.4f}')
        res_micro = np.round(float(a.grad),4)
        # test 
        self.assertAlmostEqual(res_robin, res_micro)

    def test_matmul(self):
        # robin 
        a = Tensor.ones((2,3), requires_grad=True)
        b = Tensor.full((3,2), 3., requires_grad=True)
        c = a @ b
        loss = c.sum()
        loss.backward()
        res_robin_1 = a.grad.tolist()
        res_robin_2 = b.grad.tolist()
        # torch 
        a = torch.ones((2,3), requires_grad=True)
        b = torch.full((3,2), 3., requires_grad=True)
        c = a @ b
        loss = c.sum()
        loss.backward()
        res_torch_1 = a.grad.numpy().tolist()
        res_torch_2 = b.grad.numpy().tolist()
        # test 
        self.assertEqual(res_robin_1, res_torch_1)
        self.assertEqual(res_robin_2, res_torch_2)

    def test_reshape_matmul(self):
        # robin 
        a = Tensor.ones((3,2), requires_grad=True)
        aa = a.reshape((2,3))
        b = Tensor.full((3,2), 3., requires_grad=True)
        c = aa @ b
        loss = c.sum()
        loss.backward()
        res_robin_1 = a.grad.tolist()
        # torch 
        a = torch.ones((3,2), requires_grad=True)
        aa = a.reshape((2,3))
        b = torch.full((3,2), 3., requires_grad=True)
        c = aa @ b
        loss = c.sum()
        loss.backward()
        res_torch_1 = a.grad.numpy().tolist()
        # test
        self.assertEqual(res_robin_1, res_torch_1)

    def test_slice(self):
        #robin
        a = Tensor.ones((3,3), requires_grad=True)
        aa = a[0:2]
        b = Tensor.full((3,2), 3., requires_grad=True)
        c = aa @ b
        loss = c.sum()
        loss.backward()
        res_robin_1 = a.grad.tolist()
        res_robin_2 = b.grad.tolist()
        # torch
        a = torch.ones((3,3), requires_grad=True)
        aa = a[0:2]
        b = torch.full((3,2), 3., requires_grad=True)
        c = aa @ b
        loss = c.sum()
        loss.backward()
        res_torch_1 = a.grad.numpy().tolist()
        res_torch_2 = b.grad.numpy().tolist()
        #test
        self.assertEqual(res_robin_1, res_torch_1)
        self.assertEqual(res_robin_2, res_torch_2)

    def test_transpose(self):
        #robin
        a = Tensor.ones((3,2), requires_grad=True)
        b = a.T
        c = Tensor.full((3,2), 3., requires_grad=True)
        d = b @ c
        loss = d.sum()
        loss.backward()
        res_robin_1 = a.grad.tolist()
        res_robin_2 = c.grad.tolist()
        # torch
        a = torch.ones((3,2), requires_grad=True)
        b = torch.transpose(a,1,0)
        c = torch.full((3,2), 3., requires_grad=True)
        d = b @ c
        loss = d.sum()
        loss.backward()
        res_torch_1 = a.grad.numpy().tolist()
        res_torch_2 = c.grad.numpy().tolist()
        #test
        self.assertEqual(res_robin_1, res_torch_1)
        self.assertEqual(res_robin_2, res_torch_2)

    def test_mean(self):
        # robin
        a = Tensor.eye(3, requires_grad=True)
        b = Tensor.full((3,3), 3., requires_grad=True)
        c = a @ b
        d = c.mean(axis=0, keepdim=True)
        e = Tensor.full((1,3), 4., requires_grad=True)
        f = d * e
        loss = f.sum()
        loss.backward()
        res_robin_1 = a.grad.tolist()
        # torch 
        a = torch.eye(3, requires_grad=True)
        b = torch.full((3,3), 3., requires_grad=True)
        c = a @ b
        d = c.mean(axis=0, keepdim=True)
        e = torch.full((1,3), 4., requires_grad=True)
        f = d * e
        loss = f.sum()
        loss.backward()
        res_torch_1 = a.grad.numpy().tolist()
        # test 
        self.assertAlmostEqual(res_robin_1, res_torch_1)

    def test_new_add(self):
        # robin 1
        a = Tensor.full((5,1), 3., requires_grad=True)
        b = Tensor.full((1,1), 2., requires_grad=True)
        c = a + b
        loss = c.mean()
        loss.backward()
        res_robin_1 = a.grad.tolist()
        res_robin_2 = b.grad.tolist()
        # torch 1
        a = torch.full((5,1), 3., requires_grad=True)
        b = torch.full((1,1), 2., requires_grad=True)
        c = a + b
        loss = c.mean()
        loss.backward()
        res_torch_1 = a.grad.tolist()
        res_torch_2 = b.grad.tolist()
        #test 1
        self.assertEqual(res_robin_1, res_torch_1)
        self.assertEqual(res_robin_2, res_torch_2)
        # robin 2
        a = Tensor.full((5,1), 3., requires_grad=True)
        b = Tensor.full((1,1), 2., requires_grad=True)
        c =  b + a
        loss = c.mean()
        loss.backward()
        res_robin_3 = a.grad.tolist()
        res_robin_4 = b.grad.tolist()
        # torch 2
        a = torch.full((5,1), 3., requires_grad=True)
        b = torch.full((1,1), 2., requires_grad=True)
        c = b + a
        loss = c.mean()
        loss.backward()
        res_torch_3 = a.grad.tolist()
        res_torch_4 = b.grad.tolist()
        #test 2
        self.assertEqual(res_robin_3, res_torch_3)
        self.assertEqual(res_robin_4, res_torch_4)

    def test_flatten(self):
        # robin 
        a = Tensor.ones((2,3), requires_grad=True)
        b = Tensor.full((3,2), 3., requires_grad=True)
        c = a @ b
        d = c.flatten()
        loss = d.mean()
        loss.backward()
        res_robin_1 = a.grad.tolist()
        res_robin_2 = b.grad.tolist()
        # torch
        a = torch.ones((2,3), requires_grad=True)
        b = torch.full((3,2), 3., requires_grad=True)
        c = a @ b
        d = c.flatten()
        loss = d.mean()
        loss.backward()
        res_torch_1 = a.grad.numpy().tolist()
        res_torch_2 = b.grad.numpy().tolist()
        # test 
        self.assertEqual(res_robin_1, res_torch_1)
        self.assertEqual(res_robin_2, res_torch_2)

    def test_new_mul(self):
        # robin 1
        a = Tensor.full((5,1), 3., requires_grad=True)
        b = Tensor.ones((1,1), requires_grad=True)
        c = a * b
        loss = c.sum()
        loss.backward()
        res_robin_1 = a.grad.tolist()
        res_robin_2 = b.grad.tolist()
        # torch 1
        a = torch.full((5,1), 3., requires_grad=True)
        b = torch.ones((1,1), requires_grad=True)
        c = a * b
        loss = c.sum()
        loss.backward()
        res_torch_1 = a.grad.numpy().tolist()
        res_torch_2 = b.grad.numpy().tolist()
        # test 1
        self.assertEqual(res_robin_1, res_torch_1)
        self.assertEqual(res_robin_2, res_torch_2)
        # robin 2
        a = Tensor.full((5,1), 3., requires_grad=True)
        b = Tensor.ones((1,1), requires_grad=True)
        c = b * a
        loss = c.sum()
        loss.backward()
        res_robin_3 = a.grad.tolist()
        res_robin_4 = b.grad.tolist()
        # torch 2
        a = torch.full((5,1), 3., requires_grad=True)
        b = torch.ones((1,1), requires_grad=True)
        c = b * a
        loss = c.sum()
        loss.backward()
        res_torch_3 = a.grad.numpy().tolist()
        res_torch_4 = b.grad.numpy().tolist()
        # test 2
        self.assertEqual(res_robin_3, res_torch_3)
        self.assertEqual(res_robin_4, res_torch_4)



from .tensor import Tensor
from typing import List, Dict
from collections import OrderedDict

# Tinygrad is the best framework ever.
def get_state_dict(obj, prefix:str='', tensor_type=Tensor) -> Dict[str, Tensor]:
  """
  Recursively converts an object and its nested elements into a state dictionary.

  Args:
      obj: The object to convert into a state dictionary.
      prefix (str): Optional. The prefix to prepend to each key in the state dictionary.
      tensor_type: Optional. The tensor type to use in the state dictionary.

  Returns:
      Dict[str, Tensor]: The state dictionary representing the object and its nested elements.
  """
  if isinstance(obj, tensor_type): return {prefix.strip('.'):obj}
  if hasattr(obj, '_asdict'): return get_state_dict(obj._asdict(), prefix, tensor_type)  # namedtuple
  if isinstance(obj, OrderedDict): return get_state_dict(dict(obj), prefix, tensor_type)
  if hasattr(obj, '__dict__'): return get_state_dict(obj.__dict__, prefix, tensor_type)
  state_dict = {}
  if isinstance(obj, (list, tuple)):
    for i,x in enumerate(obj): state_dict.update(get_state_dict(x, f"{prefix}{str(i)}.", tensor_type))
  elif isinstance(obj, dict):
    for k,v in obj.items(): state_dict.update(get_state_dict(v, f"{prefix}{str(k)}.", tensor_type))
  return state_dict

def get_parameters(obj) -> List[Tensor]: return list(get_state_dict(obj).values())